#!/usr/bin/python
###########################################
## TweetNLP.py
##
## interface with CMU twitter tokenizer / pos tagger
##
## ./runTagger.sh -input example_tweets.txt -output tagged_tweets.txt
##
## hansens@seas.upenn.edu

import argparse
from subprocess import check_call, Popen, PIPE
from pprint import pprint
import os

from ..dlaConstants import DEF_TOOLS_PATH

_DefaultParams ={
    'tagger_dir' : DEF_TOOLS_PATH + '/twittertagger/ark-tweet-nlp',
    'tagger_command' : 'runTagger.sh',
    'tagger_args' : ['--input-format', 'text', '--output-format', 'pretsv'],
    'tokenizer_command' : 'twokenize.sh',
    'tokenizer_args' : ['--input-format', 'text', '--output-format', 'pretsv'],
    };

class TweetNLP:

    def __init__(self, **kwargs):
        self.__dict__.update(_DefaultParams)
        self.__dict__.update(kwargs)
        if not isinstance(self.tagger_args, list):
            self.tagger_args = eval(self.tagger_args)
        if not isinstance(self.tokenizer_args, list):
            self.tokenizer_args = eval(self.tokenizer_args)
        self.taggerP = None
        self.tokenizerP = None


    ###TAGGER###

    def getTaggerProcess(self):
        #start the server:
        if not self.taggerP:
            command = self.tagger_dir+'/'+self.tagger_command
            self.taggerP =  Popen([command]+self.tagger_args, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        return self.taggerP

        
    def tag(self, sents):
        """returns a list of lists of tuples (word, tag) """
        """List is in the same order as the sents"""
        if sents:
            givenList = isinstance(sents, list)
            if not givenList: sents = [sents]

            taggerP = self.getTaggerProcess()
            taglists = []

            #call the tagger on each input:
            for s in sents:
                if s:
                    s = s.strip().replace("\n", '  ') + "\n"
                    taggerP.stdin.write(s.encode('utf-8'))
                    taggerP.stdin.flush()
                    tagLine = taggerP.stdout.readline().strip()
                    if tagLine:
                        info = tagLine.decode().split('\t')
                        (tokens, tags, probs) =  [l.split() for l in info[:3]]
                        taglists.append({'tokens':tokens, 'tags':tags, 'probs':probs, 'original':info[3]})
                else:
                        taglists.append({'tokens':[], 'tags':[], 'probs':[], 'original':s})

            if givenList:
                return taglists
            else:
                if taglists:
                    return taglists[0]
                return None
        return None


    ###TOKENIZER###

    def getTokenizerProcess(self):
        #start the server:
        if not self.tokenizerP:
            command = self.tagger_dir+'/'+self.tokenizer_command
            self.tokenizerP =  Popen([command]+self.tokenizer_args, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        return self.tokenizerP

        
    def tokenize(self, sents):
        """returns a list of lists of tuples (word, tag) """
        """List is in the same order as the sents"""
        givenList = isinstance(sents, list)
        if not givenList: sents = [sents]
        
        tokenizerP = self.getTokenizerProcess()
        tokenizedLists = []

        #call the tokenizer on each input:
        for s in sents:
            s = s.strip().replace("\n", '  ') + "\n"
            tokenizerP.stdin.write(s.encode('utf8'))
            tokenizerP.stdin.flush()
            tagLine = tokenizerP.stdout.readline().strip()
            if tagLine:
                info = tagLine.decode().split('\t')
                tokens =  info[0].split()
                tokenizedLists.append(tokens)

        if givenList:
            return tokenizedLists
        else:
            return tokenizedLists[0]



if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='Test the TweetNLP python interface.', prefix_chars='-+', formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    for param, value in _DefaultParams.items():
        parser.add_argument('--'+str(param), metavar='string', dest=str(param), default=str(value),
                        help="%s default param" % param)

    parser.add_argument('-t', '--tag', metavar='string', dest='taglines', nargs='+', default=[],
                        help="lines to parse (place each in quotes)")

    parser.add_argument('-o', '--tokenize', metavar='string', dest='tokenlines', nargs='+', default=[],
                        help="lines to parse (place each in quotes)")

    args = parser.parse_args()

    if args.taglines:
        pprint(TweetNLP(**args.__dict__).tag(args.taglines))
    if args.tokenlines:
        pprint(TweetNLP(**args.__dict__).tokenize(args.tokenlines))




import re
import sys
from html.parser import HTMLParser
from pprint import pprint

#infrastructure
from .dlaWorker import DLAWorker
from . import textCleaner as tc
from . import dlaConstants as dlac
from .mysqlmethods import mysqlMethods as mm
from .lib.happierfuntokenizing import Tokenizer #Potts tokenizer

try:
    from langid.langid import LanguageIdentifier, model
except ImportError:
    dlac.warn("Cannot import langid (cannot use addLanguageFilterTable)")
    pass

class MessageAnnotator(DLAWorker):
    """Deals with filtering or adding columns to message tables.

    Returns
    -------
    MessageAnnotator object

    Examples
    --------

    """

    def addDedupFilterTable(self, anonymize=True):
        """
        Groups all messages in a given table and filters deplicate messages within the correl_field. Writes
        a new message table called corptable_dedup. Deduplication written by Daniel Preotiuc-Pietro and adapted for DLATK.
        Removal of urls, punctuation, etc. taken from the twokenize tokenizer written by Brendan O'connor.
        """
        tokenizer = Tokenizer(use_unicode=self.use_unicode)

        new_table = self.corptable + "_dedup"
        drop = """DROP TABLE IF EXISTS %s""" % (new_table)
        create = """CREATE TABLE %s like %s""" % (new_table, self.corptable)
        mm.execute(self.corpdb, self.dbCursor, drop, charset=self.encoding, use_unicode=self.use_unicode)
        mm.execute(self.corpdb, self.dbCursor, create, charset=self.encoding, use_unicode=self.use_unicode)
        mm.standardizeTable(self.corpdb, self.dbCursor, new_table, collate=dlac.DEF_COLLATIONS[self.encoding.lower()], engine=dlac.DEF_MYSQL_ENGINE, charset=self.encoding, use_unicode=self.use_unicode)

        #Find column names:
        columnNames = list(mm.getTableColumnNameTypes(self.corpdb, self.dbCursor, self.corptable, charset=self.encoding, use_unicode=self.use_unicode).keys())
        messageIndex = columnNames.index(self.message_field)
        try:
            retweetedStatusIdx = columnNames.index("retweeted_status_text")
        except:
            retweetedStatusIdx = None
            pass

        #find all groups that are not already inserted
        usql = """SELECT %s FROM %s GROUP BY %s""" % (self.correl_field, self.corptable, self.correl_field)
        cfRows = [r[0] for r in mm.executeGetList(self.corpdb, self.dbCursor, usql, charset=self.encoding, use_unicode=self.use_unicode)]
        dlac.warn("deduplicating messages for %d '%s's"%(len(cfRows), self.correl_field))

        # if message level analysis
        if any(field in self.correl_field.lower() for field in ["mess", "msg"]) or self.correl_field.lower().startswith("id"):
            groupsAtTime = 1
            dlac.warn("""Deduplicating only works with a non-message level grouping such as users or counties.""", attention=True)
            sys.exit()

        groupsAtTime = 1
        rows_to_write = []
        counter = 1
        for groups in dlac.chunks(cfRows, groupsAtTime):

            # get msgs for groups:
            sql = """SELECT %s from %s where %s IN ('%s')""" % (','.join(columnNames), self.corptable, self.correl_field, "','".join(str(g) for g in groups))
            rows = list(mm.executeGetList(self.corpdb, self.dbCursor, sql, warnQuery=False, charset=self.encoding, use_unicode=self.use_unicode))
            rows = [row for row in rows if row[messageIndex] and not row[messageIndex].isspace()]

            bf = []
            for row in rows:
                row = list(row)
                try:
                    message = row[messageIndex]
                    textrt = ''
                    if retweetedStatusIdx:
                        textrt = row[retweetedStatusIdx]
                    if not textrt:
                        textrt = tc.rttext(message)[0]
                    if not textrt == '':
                        message = textrt
                        continue
                    #words = tokenizer.tokenize(message)
                    words = [word for word in tokenizer.tokenize(message) if ((word[0]!='#') and (word[0]!='@') and not tc.Exclude_RE.search(word))]

                    message = ' '.join(words).lower()
                    if len(words)>=6:
                        if 'YouTube' in words:
                            message = ' '.join(words[0:5])
                        else:
                            message = ' '.join(words[0:5])
                    if message not in bf:
                        bf.append(message)
                        if anonymize:
                            message = row[messageIndex]
                            message = tc.replaceURL(message)
                            row[messageIndex] = message
                        rows_to_write.append(row)
                    else:
                        pass
                except:
                    continue
            if len(rows_to_write) >= dlac.MYSQL_BATCH_INSERT_SIZE:
                sql = """INSERT INTO """+new_table+""" ("""+', '.join(columnNames)+""") VALUES ("""  +", ".join(['%s']*len(columnNames)) + """)"""
                mm.executeWriteMany(self.corpdb, self.dbCursor, sql, rows_to_write, writeCursor=self.dbConn.cursor(), charset=self.encoding, use_unicode=self.use_unicode)
                rows_to_write = []

            if (counter % 500 == 0):
                print('%d deduplicated users inserted!' % (counter))
            counter += 1

        if rows_to_write:
            sql = """INSERT INTO """+new_table+""" ("""+', '.join(columnNames)+""") VALUES ("""  +", ".join(['%s']*len(columnNames)) + """)"""
            mm.executeWriteMany(self.corpdb, self.dbCursor, sql, rows_to_write, writeCursor=self.dbConn.cursor(), charset=self.encoding, use_unicode=self.use_unicode)

    def addAnonymizedTable(self):
        """
        
        """
        new_table = self.corptable + "_an"
        drop = """DROP TABLE IF EXISTS %s""" % (new_table)
        create = """CREATE TABLE %s like %s""" % (new_table, self.corptable)
        mm.execute(self.corpdb, self.dbCursor, drop, charset=self.encoding, use_unicode=self.use_unicode)
        mm.execute(self.corpdb, self.dbCursor, create, charset=self.encoding, use_unicode=self.use_unicode)
        mm.standardizeTable(self.corpdb, self.dbCursor, new_table, collate=dlac.DEF_COLLATIONS[self.encoding.lower()], engine=dlac.DEF_MYSQL_ENGINE, charset=self.encoding, use_unicode=self.use_unicode)

        #Find column names:
        columnNames = list(mm.getTableColumnNameTypes(self.corpdb, self.dbCursor, self.corptable, charset=self.encoding, use_unicode=self.use_unicode).keys())
        messageIndex = columnNames.index(self.message_field)
        try:
            retweetedStatusIdx = columnNames.index("retweeted_status_text")
        except:
            retweetedStatusIdx = None
            pass

        #find all groups that are not already inserted
        usql = """SELECT %s FROM %s GROUP BY %s""" % (self.correl_field, self.corptable, self.correl_field)
        cfRows = [r[0] for r in mm.executeGetList(self.corpdb, self.dbCursor, usql, charset=self.encoding, use_unicode=self.use_unicode)]
        dlac.warn("anonymizing messages for %d '%s's"%(len(cfRows), self.correl_field))

        groupsAtTime = 1
        rows_to_write = []
        counter = 1
        for groups in dlac.chunks(cfRows, groupsAtTime):

            # get msgs for groups:
            sql = """SELECT %s from %s where %s IN ('%s')""" % (','.join(columnNames), self.corptable, self.correl_field, "','".join(str(g) for g in groups))
            rows = list(mm.executeGetList(self.corpdb, self.dbCursor, sql, warnQuery=False, charset=self.encoding, use_unicode=self.use_unicode))
            rows = [row for row in rows if row[messageIndex] and not row[messageIndex].isspace()]

            for row in rows:
                row = list(row)
                try:
                    message = row[messageIndex]
                    message = tc.replaceUser(message)
                    message = tc.replaceURL(message)
                    row[messageIndex] = message
                    rows_to_write.append(row)
                except:
                    continue
            if len(rows_to_write) >= dlac.MYSQL_BATCH_INSERT_SIZE:
                sql = """INSERT INTO """+new_table+""" ("""+', '.join(columnNames)+""") VALUES ("""  +", ".join(['%s']*len(columnNames)) + """)"""
                mm.executeWriteMany(self.corpdb, self.dbCursor, sql, rows_to_write, writeCursor=self.dbConn.cursor(), charset=self.encoding, use_unicode=self.use_unicode)
                rows_to_write = []

            if (counter % 500 == 0):
                print('%d anonymized messages inserted!' % (counter))
            counter += 1

        if rows_to_write:
            sql = """INSERT INTO """+new_table+""" ("""+', '.join(columnNames)+""") VALUES ("""  +", ".join(['%s']*len(columnNames)) + """)"""
            mm.executeWriteMany(self.corpdb, self.dbCursor, sql, rows_to_write, writeCursor=self.dbConn.cursor(), charset=self.encoding, use_unicode=self.use_unicode)


    def addSpamFilterTable(self, threshold=dlac.DEF_SPAM_FILTER):
        """
        Groups all messages in a given table and filters spam messages within the correl_field. Writes
        a new message table called corptable_nospam with additional integer column is_spam (0 = not spam, 1 = spam).
        Spam words = 'share', 'win', 'check', 'enter', 'products', 'awesome', 'prize', 'sweeps', 'bonus', 'gift'

        Parameters
        ----------
        threshold : float
            percentage of spam messages to be considered a spam user
        """
        spam_words = ['share', 'win', 'check', 'enter', 'products', 'awesome', 'prize', 'sweeps', 'bonus', 'gift']

        new_table = self.corptable + "_nospam"
        drop = """DROP TABLE IF EXISTS %s""" % (new_table)
        create = """CREATE TABLE %s like %s""" % (new_table, self.corptable)
        add_colum = """ALTER TABLE %s ADD COLUMN is_spam INT(2) NULL""" % (new_table)
        mm.execute(self.corpdb, self.dbCursor, drop, charset=self.encoding, use_unicode=self.use_unicode)
        mm.execute(self.corpdb, self.dbCursor, create, charset=self.encoding, use_unicode=self.use_unicode)
        mm.execute(self.corpdb, self.dbCursor, add_colum, charset=self.encoding, use_unicode=self.use_unicode)
        mm.standardizeTable(self.corpdb, self.dbCursor, new_table, collate=dlac.DEF_COLLATIONS[self.encoding.lower()], engine=dlac.DEF_MYSQL_ENGINE, charset=self.encoding, use_unicode=self.use_unicode)

        #Find column names:
        columnNames = list(mm.getTableColumnNameTypes(self.corpdb, self.dbCursor, self.corptable, charset=self.encoding, use_unicode=self.use_unicode).keys())
        insertColumnNames = columnNames + ['is_spam']
        messageIndex = columnNames.index(self.message_field)

        #find all groups that are not already inserted
        usql = """SELECT %s FROM %s GROUP BY %s""" % (self.correl_field, self.corptable, self.correl_field)
        cfRows = [r[0] for r in mm.executeGetList(self.corpdb, self.dbCursor, usql, charset=self.encoding, use_unicode=self.use_unicode)]
        dlac.warn("Removing spam messages for %d '%s's"%(len(cfRows), self.correl_field))

        # if message level analysis
        if any(field in self.correl_field.lower() for field in ["mess", "msg"]) or self.correl_field.lower().startswith("id"):
            groupsAtTime = 1
            dlac.warn("""This will remove any messages that contain *any* spam words. Consider rerunning at the user level.""", attention=True)


        groupsAtTime = 1
        rows_to_write = []
        counter = 1
        users_removed = 0
        for groups in dlac.chunks(cfRows, groupsAtTime):

            # get msgs for groups:
            sql = """SELECT %s from %s where %s IN ('%s')""" % (','.join(columnNames), self.corptable, self.correl_field, "','".join(str(g) for g in groups))
            rows = list(mm.executeGetList(self.corpdb, self.dbCursor, sql, warnQuery=False, charset=self.encoding, use_unicode=self.use_unicode))
            rows = [row for row in rows if row[messageIndex] and not row[messageIndex].isspace()]

            total_messages = float(len(rows))
            if total_messages == 0: continue
            spam_messages = 0
            insert_rows = []
            for row in rows:
                try:
                    if any(word in row[messageIndex] for word in spam_words):
                        spam_messages += 1
                        insert_rows.append(row + (1,))
                    else:
                        insert_rows.append(row + (0,))
                except:
                    continue

            if spam_messages/total_messages < threshold:
                rows_to_write += insert_rows
            else:
                users_removed += 1

            if len(rows_to_write) >= dlac.MYSQL_BATCH_INSERT_SIZE:
                sql = """INSERT INTO """+new_table+""" ("""+', '.join(insertColumnNames)+""") VALUES ("""  +", ".join(['%s']*len(insertColumnNames)) + """)"""
                mm.executeWriteMany(self.corpdb, self.dbCursor, sql, rows_to_write, writeCursor=self.dbConn.cursor(), charset=self.encoding, use_unicode=self.use_unicode)
                rows_to_write = []

            if (counter % 500 == 0):
                print('%d users filtered!' % (counter))
            counter += 1

        if rows_to_write:
            sql = """INSERT INTO """+new_table+""" ("""+', '.join(insertColumnNames)+""") VALUES ("""  +", ".join(['%s']*len(insertColumnNames)) + """)"""
            mm.executeWriteMany(self.corpdb, self.dbCursor, sql, rows_to_write, writeCursor=self.dbConn.cursor(), charset=self.encoding, use_unicode=self.use_unicode)
        print('%d users removed!' % (users_removed))

    # TODO: add nicer implementation
    def yieldMessages(self, messageTable, totalcount):
        if totalcount > 10*dlac.MAX_SQL_SELECT:
            for i in range(0,totalcount, dlac.MAX_SQL_SELECT):
                sql = "SELECT * FROM %s limit %d, %d" % (messageTable, i, dlac.MAX_SQL_SELECT)
                for m in mm.executeGetList(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode):
                    yield [i for i in m]
        else:
            sql = "SELECT * FROM %s" % messageTable
            for m in mm.executeGetList(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode):
                yield m

    def addLanguageFilterTable(self, langs, cleanMessages, lowercase):
        """Filters all messages in corptable for a given language. Keeps messages if
        confidence is greater than 80%. Uses the langid library.

        Parameters
        ----------
        langs : list
            list of languages to filter for
        cleanMessages : boolean
            remove URLs, hashtags and @ mentions from messages before running langid
        lowercase : boolean
            convert message to all lowercase before running langid
        """
        identifier = LanguageIdentifier.from_modelstring(model, norm_probs=True)

        new_table = self.corptable + "_%s"

        columnNames = mm.getTableColumnNames(self.corpdb, self.corptable, charset=self.encoding, use_unicode=self.use_unicode)
        messageIndex = [i for i, col in enumerate(columnNames) if col.lower() == dlac.DEF_MESSAGE_FIELD.lower()][0]
        #messageIDindex = [i for i, col in enumerate(columnNames) if col.lower() == dlac.DEF_MESSAGEID_FIELD.lower()][0]

        # CREATE NEW TABLES IF NEEDED
        messageTables = {l: new_table % l for l in langs}
        for l, table in messageTables.items():
            drop = """DROP TABLE IF EXISTS %s""" % (table)
            create = """CREATE TABLE %s like %s""" % (table, self.corptable)
            mm.execute(self.corpdb, self.dbCursor, drop, charset=self.encoding, use_unicode=self.use_unicode)
            mm.execute(self.corpdb, self.dbCursor, create, charset=self.encoding, use_unicode=self.use_unicode)
            mm.standardizeTable(self.corpdb, self.dbCursor, table, collate=dlac.DEF_COLLATIONS[self.encoding.lower()], engine=dlac.DEF_MYSQL_ENGINE, charset=self.encoding, use_unicode=self.use_unicode)

        #ITERATE THROUGH EACH MESSAGE WRITING THOSE THAT ARE ENGLISH
        messageDataToAdd = {l: list() for l in langs}
        messageDataCounts = {l: 0 for l in langs}
        totalMessages = 0
        totalMessagesKept = 0
        sql = """SELECT COUNT(*) FROM %s""" % self.corptable
        totalMessagesInTable = mm.executeGetList(self.corpdb, self.dbCursor, sql)[0][0]

        print("Reading %s messages" % ",".join([str(totalMessagesInTable)[::-1][i:i+3] for i in range(0,len(str(totalMessagesInTable)),3)])[::-1])
        memory_limit = dlac.MYSQL_BATCH_INSERT_SIZE if dlac.MYSQL_BATCH_INSERT_SIZE < totalMessagesInTable else totalMessagesInTable/20

        html = HTMLParser()
        for messageRow in self.yieldMessages(self.corptable, totalMessagesInTable):
            messageRow = list(messageRow)
            totalMessages+=1
            message = messageRow[messageIndex]

            try:
                message = html.unescape(message)
                messageRow[messageIndex] = message

                if cleanMessages:
                    message = re.sub(r"(?:\#|\@|https?\://)\S+", "", message)
            except Exception as e:
                print(e)
                print([message])

            try:
                if lowercase:
                    message = message.lower()
            except Exception as e:
                print(e)
                print([message])


            lang, conf = None, None
            try:
                lang, conf = identifier.classify(message)
            except TypeError as e:
                print(("         Error, ignoring row %s" % str(messageRow)))

            if lang in langs and conf > dlac.DEF_LANG_FILTER_CONF:
                messageDataToAdd[lang].append(messageRow)
                messageDataCounts[lang] += 1

                totalMessagesKept+=1
            else:
                continue

            if totalMessagesKept % memory_limit == 0:
                #write messages every so often to clear memory
                for l, messageData in messageDataToAdd.items():
                    sql = """INSERT INTO """+messageTables[l]+""" ("""+', '.join(columnNames)+""") VALUES ("""  +", ".join(['%s']*len(columnNames)) + """)"""
                    mm.executeWriteMany(self.corpdb, self.dbCursor, sql, messageData, writeCursor=self.dbConn.cursor(), charset=self.encoding, use_unicode=self.use_unicode)
                    messageDataToAdd[l] = list()

                for l, nb in [(x[0], len(x[1])) for x in iter(messageDataToAdd.items())]:
                    messageDataCounts[l] += nb

                print("  %6d rows written (%6.3f %% done)" % (totalMessagesKept,
                                                              100*float(totalMessages)/totalMessagesInTable))

        if messageDataToAdd:
            print("Adding final rows")
            for l, messageData in messageDataToAdd.items():
                sql = """INSERT INTO """+messageTables[l]+""" ("""+', '.join(columnNames)+""") VALUES ("""  +", ".join(['%s']*len(columnNames)) + """)"""
                mm.executeWriteMany(self.corpdb, self.dbCursor, sql, messageData, writeCursor=self.dbConn.cursor(), charset=self.encoding, use_unicode=self.use_unicode)

        print("Kept %d out of %d messages" % (totalMessagesKept, totalMessages))
        pprint({messageTables[l]: v for l, v in messageDataCounts.items()})



"""
MySQL interface methods based on the SQLAlchemy package
"""

import socket
import logging
import itertools
from sqlalchemy import create_engine
from sqlalchemy.engine.url import URL
import pandas as pd
import warnings
import sys
import time
import MySQLdb.cursors

from dlatk.dlaConstants import DEF_ENCODING, MYSQL_ERROR_SLEEP, MYSQL_HOST, MAX_ATTEMPTS, warn

def get_db_engine(db_schema, db_host = MYSQL_HOST, charset=DEF_ENCODING, db_config = '~/.my.cnf', port=3306, stream=False):
    eng = None
    attempts = 0;
    while (1):
        try:
            db_url = URL(drivername='mysql', host=db_host, port=port,
                database=db_schema,
                query={
                    'read_default_file' : db_config,
                    'charset': charset
                })
            if stream:
                eng = create_engine(name_or_url=db_url, connect_args={'cursorclass': MySQLdb.cursors.SSCursor}, pool_recycle=600)
            else:
                eng = create_engine(name_or_url=db_url)
            break
        except Exception as e:
            attempts += 1
            warn(" *MYSQL Connect ERROR on db:%s\n%s\n (%d attempt)"% (db, e, attempts))
            time.sleep(MYSQL_ERROR_SLEEP*attempts**2)
            if (attempts > MAX_ATTEMPTS):
                sys.exit(1)
    return eng

def create_table(db_eng, table_name, columns, more_definitions=[], table_options="", if_exists="error"):
    '''
    :param db_eng: sqlalchemy.Engine
    :param table_name: string
    :param columns: OrderedDict or dict - where keys are column names and values are column type definitions
    :param more_details: list
    :param table_options: string
    :param if_exists: string ('error' or 'replace' or 'skip')
    :return:
    '''
    res = db_eng.execute("SHOW TABLES LIKE '{}'".format(table_name))
    table_exists = res.rowcount == 1
    if table_exists and if_exists.lower() == "replace":
        db_eng.execute("DROP TABLE IF EXISTS {}".format(table_name))
    if not (table_exists and (if_exists.lower() == "skip")):
        assert type(list(columns.values())[0]) is str, "Values for column dict must be strings."
        if type(more_definitions) is not list:
            more_definitions = [more_definitions]
        table_def_string = ', '.join(["{} {}".format(column_name, columns[column_name]) for column_name in columns] + more_definitions)
        db_eng.execute("CREATE TABLE {} ({}) {}".format(table_name, table_def_string, table_options))

def extend_table(db_eng, table_name, new_columns, if_exists="skip"):
    '''
    Add columns to a table if they don't already exist
    :param new_columns: OrderedDict or dict - where keys are column names and values are column type definitions
    :param if_exists: string('skip') #no other options implemented yet
    '''
    if if_exists != "skip":
        raise NotImplementedError
    db_schema = db_eng.url.database
    rows = db_eng.execute("SELECT * FROM information_schema.COLUMNS WHERE TABLE_SCHEMA = '%s' AND TABLE_NAME = '%s' AND COLUMN_NAME = '%s'" % (db_schema, table_name, list(new_columns.keys())[0]))
    if rows.rowcount == 0:
        print("Adding new columns to table {}...".format(table_name))
        add_column_phrase = ", ".join(["ADD COLUMN {} {}".format(column_name, new_columns[column_name]) for column_name in new_columns])
        db_eng.execute("ALTER TABLE {} {}".format(table_name, add_column_phrase))
        print("Done.")
    else:
        print("Some or all of the columns exist. Ignoring.")
        for row in rows:
            print(row)

def list_iter_to_dict_iter(list_iter, dict_keys):
    '''Take list iter + keys, return dict iter'''
    for mylist in list_iter:
        if type(mylist) is not list:
            warnings.warn("list_iter_to_dict is not getting lists, it is getting {}".format(str(type(mylist))))
        yield dict(list(zip(dict_keys, mylist)))

def resultset_to_dict_iter(result_set):
    '''
    :param result_set: SQLAlchemy results set from db_eng.execute(sql)
    :return: iterator of dicts!
    '''
    for rp in result_set:
        yield dict(list(rp.items()))

def select_columns(dict_iter, keys_to_keep):
    '''Take a dict iter, return a dict iter with fewer columns'''
    for mydict in dict_iter:
        yield dict((k, mydict[k]) for k in keys_to_keep)

def rename_keys(dict_iter, renamings):
    for mydict in dict_iter:
        for old_key in renamings:
            new_key = renamings[old_key]
            mydict[new_key] = mydict[old_key]
            del mydict[old_key]
        yield mydict

def mysql_insert(db_eng, db_table, dict_iterator):
    '''
    :param dict_iterator: iterator where next() yields dict that corresponds to a row of data
    :param db_eng: SQLAlchemy engine
    :param db_table: expects this to already exist
    steps through iterator, and inserts each yielded item into a mysql database
    '''
    first_row = next(dict_iterator)

    list_of_columns = list(first_row.keys())
    columns = ', '.join(list_of_columns)
    num_columns = len(list_of_columns)
    wildcards = ','.join(['%s']*num_columns)
    insert_statement = "INSERT INTO {0} ({1}) values({2})".format(db_table, columns, wildcards)

    _mysql_insert_row(db_eng, insert_statement, first_row, list_of_columns)
    for row_dict in dict_iterator:
        _mysql_insert_row(db_eng, insert_statement, row_dict, list_of_columns)

def _mysql_insert_row(db_eng, insert_statement, values_dict, list_of_columns):
    '''Helper function for mysql_insert'''
    values = [values_dict[column] for column in list_of_columns]
    try:
        db_eng.execute(insert_statement, values)
    except Exception as e:
        logging.warning("Error inserting data: {}{}".format(type(e), e))

def mysql_update(db_eng, db_table, dict_iterator, unique_column="id", log_every=10000):
    '''
    Updates a database table with values defined in dict iterator, finds rows using unique column
    Other fields are assumed to have the same name as their dict key
    :param unique_column: the column name for unique column, probably the id field
    '''
    first_row = next(dict_iterator)
    update_columns = list(first_row.keys())
    del update_columns[update_columns.index(unique_column)]
    update_statements = ", ".join(["{} = %s".format(col) for col in update_columns])
    update_sql = "UPDATE {} SET {} WHERE {} = %s".format(db_table,update_statements,unique_column)
    ###
    _mysql_update_row(db_eng, update_sql, first_row, unique_column, update_columns)
    count = 0
    starttime = time.time()
    for row in dict_iterator:
        count += 1
        if count % log_every == 0:
            time_elapsed = time.time() - starttime
            print("{} rows updated in {}s...".format(count, time_elapsed))
        _mysql_update_row(db_eng, update_sql, row, unique_column, update_columns)

def _mysql_update_row(db_eng, update_sql, row, unique_column, update_columns):
    '''Helper function for mysql_update'''
    unique_value = row[unique_column]
    #TODO - give a warning when row[column] is missing
    #warnings.warn("update data missing")
    update_values = [row[column] if column in row else None for column in update_columns]
    values = update_values + [unique_value]
    try:
        db_eng.execute(update_sql, values)
    except Exception as e:
        logging.warning("Error updating data: {}{}".format(type(e), e))

def mysql_multitable(db_eng, dict_iter, table_prefix, table_column, table_column_transform, batch_size=1000, template_table=None ):
    '''
    Takes an iterator and outputs into multiple tables based on data from the iterator
    :param db_eng: SQLAlchemy db engine
    :param dict_iter: an iterator of dicts
    :param table_prefix: for the naming of the tables
    :param table_column: column to use for deciding which table to insert into
    :param table_column_transform: function to generate table name suffix using 'table_column' value
    :param batch_size: once a target table is identified, assume the next n rows go here
    '''

    batch = list(itertools.islice(dict_iter, 0, batch_size))
    batch_num = 0
    while len(batch) > 0:
        df_batch = pd.DataFrame(batch)
        df_batch["batch_num"] = batch_num
        first_line = dict(df_batch.ix[0])
        table_name = table_prefix + table_column_transform(first_line[table_column])
        try:
            if template_table:
                db_eng.execute("CREATE TABLE IF NOT EXISTS {} LIKE {}".format(table_name, template_table))
            df_batch.to_sql(table_name, db_eng, if_exists="append", index=False)
        except Exception as e:
            exc_info = sys.exc_info()
            print("Unexpected error:", exc_info[0], exc_info[1], exc_info[2])
        else:
            print("Successful insert of {} rows!".format(len(df_batch)))
        batch = list(itertools.islice(dict_iter, 0, batch_size))
        batch_num += 1


def dictify(my_iter, columns):
    for item in my_iter:
        yield dict(list(zip(columns, item)))



#!/usr/bin/python
###########################################
## StanfordParser.py
##
## requires the following to be in a shell script, usually named oneline.sh
## in the parser directory: 
##
## #!/usr/bin/env bash
## scriptdir=`dirname $0`
## java -mx750m -cp "$scriptdir/stanford-parser.jar:" edu.stanford.nlp.parser.lexparser.LexicalizedParser -sentences newline -outputFormat "oneline,wordsAndTags,typedDependenciesCollapsed" $scriptdir/grammar/englishPCFG.ser.gz $*

import sys
import os
import argparse
import subprocess
import random
import re
from pprint import pprint
import glob

from ..dlaConstants import warn, DEF_STANFORD_PARSER

_DefaultParams ={
    'save_file' : 'parsed.data',
    'save_dir' : 'backupParses',
    'parser_dir' : DEF_STANFORD_PARSER,
    'parser_command' : 'oneline.sh', #note this was an edited lexparser.csh to include oneline output option to make the tree easier to handle
    'temp_file' : DEF_STANFORD_PARSER + '/temp.file', #holds sentences to be parsed
    'max_sent_words' : int(60),
    #'split' : 0,#whether to ask the parse to split sentences.(NOTE: when on, also sets a maximum sentence size)
    };

class StanfordParser:

    def __init__(self, **kwargs):
        self.__dict__.update(_DefaultParams)
        self.__dict__.update(kwargs)
        if not self.parser_dir:
            warn("Cannot find StanfordParser, please check that this was installed properly")
            sys.exit()
        
    def parse(self, sents):
        """returns a list of dicts with parse information: const, dep, and pos """
        """List is in the same order as the sents"""

        #create temp fileName
        tempFileName = self.temp_file+'.'+str(random.randint(1, 1000000))
        
        #write the sents to a file
        f = open(tempFileName,'w')
        for s in sents:
            s = shortenToNWords(s.strip().replace("\n", ' '), int(self.max_sent_words))
            f.write(s)
            f.write("\n")
        f.close()

        #call the parser
        command = self.parser_dir+'/'+self.parser_command
        p = subprocess.Popen([command, tempFileName], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        output, errors = p.communicate()

        #parse the results
        #print output #debug
        parseDicts = self.getParseDicts(output)
        if not len(parseDicts) == len(sents):
            for i in range(min(len(parseDicts), len(sents))):
                pprint((sents[i], parseDicts[i]['pos']))
                print("\n")
            print("Number of parses does not match number of sents")
            print(" number of sents:  %d" % len(sents))
            print(" number of parses: %d" % len(parseDicts))
            print(sys.stderr.write("!!ERROR!! sents and parses do not match")) 
            sys.exit(0)

        os.remove(tempFileName)

        return parseDicts

    depParseRe = re.compile(r'^[a-z0-9\_]+\([a-z0-9]', re.I)
    def getParseDicts(self, output):
        parseDicts = []
        lines = output.decode().split("\n")
        i = 0
        while i < len(lines):
            parse = dict()
            #get pos
            while i < len(lines) and not lines[i]:
                i+=1
            if i >= len(lines): break
            parse['pos'] = lines[i].strip()
            i+=1

            #get const
            while i < len(lines) and not lines[i]:
                i+=1
            if i >= len(lines): break
            parse['const'] = lines[i].strip()
            i+=1

            #get dep
            depLines = []
            while i < len(lines) and not lines[i]:
                i+=1
            while i < len(lines) and lines[i]:
                depLines.append(lines[i].strip())
                i+=1
            if i < len(lines) and len(depLines) < 2 and not StanfordParser.depParseRe.match(depLines[0]):
                i-=1
                depLines = []
            parse['dep'] = depLines

            parseDicts.append(parse)

        return parseDicts


def shortenToNWords(sent, n):
    words = sent.split()
    return ' '.join(words[:int(n)])

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='Test the stanford parser python interface.', prefix_chars='-+', formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    for param, value in _DefaultParams.items():
        parser.add_argument('--'+str(param), metavar='string', dest=str(param), default=str(value),
                        help="%s default param" % param)

    parser.add_argument('-p', '--parse', metavar='string', dest='parselines', nargs='+', default=[],
                        help="lines to parse (place each in quotes)")

    args = parser.parse_args()

    if args.parselines:
        pprint(StanfordParser(**args.__dict__).parse(args.parselines))




__authors__ = "H. Andrew Schwartz, Salvatore Giorgi, Maarten Sap, Patrick Crutchley, Lukasz Dziurzynski and Megha Agrawal"
__copyright__ = "Copyright 2017"
__credits__ = []
__license__ = "Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported License: http://creativecommons.org/licenses/by-nc-sa/3.0/"
__version__ = "1.1.4" 
__maintainer__ = "Salvatore Giorgi, Andy Schwartz"
__email__ = "sgiorgi@sas.upenn.edu, has@cs.stonybrook.edu"



import ast
import csv
import json
import time
import multiprocessing
import imp
import sys
import os
import re

#infrastructure
from .dlaWorker import DLAWorker
from . import dlaConstants as dlac
from . import textCleaner as tc
from .mysqlmethods import mysqlMethods as mm
from .lib.happierfuntokenizing import Tokenizer #Potts tokenizer
try:
    from .lib.StanfordParser import StanfordParser
except ImportError:
    dlac.warn("Cannot import StanfordParser (interface with the Stanford Parser)")
    pass

#nltk
try:
    import nltk.data
except ImportError:
    print("warning: unable to import nltk.tree or nltk.corpus or nltk.data")
try:
    from .lib.TweetNLP import TweetNLP
except ImportError:
    dlac.warn("Cannot import TweetNLP (interface with CMU Twitter tokenizer / pos tagger)")
    pass


class MessageTransformer(DLAWorker):
    """Deals with message tables .....

    Returns
    -------
    MessageTransformer object

    Examples
    --------

    """

    groupsAtTime = 100

    def __createTable(self, tableName, modify=''):
        drop = """DROP TABLE IF EXISTS %s""" % (tableName)
        mm.execute(self.corpdb, self.dbCursor, drop, charset=self.encoding, use_unicode=self.use_unicode)

        sql = """CREATE TABLE %s like %s""" % (tableName, self.corptable)
        mm.execute(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode)

        if modify:
            alter = """ALTER TABLE %s MODIFY %s %s""" % (tableName, self.message_field, modify)
            mm.execute(self.corpdb, self.dbCursor, alter, charset=self.encoding, use_unicode=self.use_unicode)
        
        mm.standardizeTable(self.corpdb, self.dbCursor, tableName, collate=dlac.DEF_COLLATIONS[self.encoding.lower()], engine=dlac.DEF_MYSQL_ENGINE, charset=self.encoding, use_unicode=self.use_unicode)
        mm.disableTableKeys(self.corpdb, self.dbCursor, tableName, charset=self.encoding, use_unicode=self.use_unicode)
    
        columnNames = list(mm.getTableColumnNameTypes(self.corpdb, self.dbCursor, self.corptable, charset=self.encoding, use_unicode=self.use_unicode).keys())
        messageIndex = columnNames.index(self.message_field)
        messageIdIndex = columnNames.index(self.messageid_field)
        return columnNames, messageIndex, messageIdIndex

    def __findAllGroups(self):
        usql = """SELECT %s FROM %s GROUP BY %s""" % (self.correl_field, self.corptable, self.correl_field)
        cfRows = [r[0] for r in mm.executeGetList(self.corpdb, self.dbCursor, usql, charset=self.encoding, use_unicode=self.use_unicode)]
        dlac.warn("finding messages for %d '%s's"%(len(cfRows), self.correl_field))
        return cfRows

    def __getMsgsForGroups(self, groups, columnNames, messageIndex):
        sql = """SELECT %s from %s where %s IN ('%s')""" % (','.join(columnNames), self.corptable, self.correl_field, "','".join(str(g) for g in groups))
        rows = list(mm.executeGetList(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode))#, False)
        return rows

    def __writeMsgsForGroups(self, rows, parses, messageIndex, tableName, columnNames):
        insert_idx_start = 0
        insert_idx_end = dlac.MYSQL_BATCH_INSERT_SIZE

        #add msgs into new tables
        sql = """INSERT INTO """+tableName+""" ("""+', '.join(columnNames)+\
                """) VALUES ("""  +", ".join(['%s']*len(columnNames)) + """)"""
        for i in range(len(rows)):
            rows[i] = list(rows[i])
            rows[i][messageIndex] = str(parses[i])

        while insert_idx_start < len(rows):
            dataToWrite = rows[insert_idx_start:min(insert_idx_end, len(rows))]
            mm.executeWriteMany(self.corpdb, self.dbCursor, sql, dataToWrite, writeCursor=self.dbConn.cursor(), charset=self.encoding, use_unicode=self.use_unicode)
            insert_idx_start += dlac.MYSQL_BATCH_INSERT_SIZE
            insert_idx_end += dlac.MYSQL_BATCH_INSERT_SIZE

    def insertLDARows(self, ldas, tableName, columnNames, messageIndex, messageIdIndex):
        """?????

        Parameters
        ----------
        ldas : ?????
            ?????
        tableName : ?????
            ?????
        columnNames : ?????
            ?????
        messageIndex : ?????
            ?????
        messageIdIndex : ?????
            ?????

        """
        message_ids = [str(msg_id) for msg_id in ldas.keys()]
        sql = """SELECT %s from %s where %s IN ('%s')""" % (
            ','.join(columnNames), self.corptable, self.messageid_field,
            "','".join(message_ids))

        rows = list(mm.executeGetList(self.corpdb, self.dbCursor, sql, False, charset=self.encoding, use_unicode=self.use_unicode))

        #generate row data:
        newRows = []
        for row in rows:
            if row[messageIndex] and not row[messageIndex].isspace():
                newRow = list(row)
                newRow[messageIndex] = json.dumps(ldas[str(row[messageIdIndex])])
                newRows.append(newRow)
                
        #insert
        sql = """INSERT INTO """+tableName+""" ("""+', '.join(columnNames)+\
            """) VALUES ("""  +", ".join(['%s']*len(columnNames)) + """)"""
        mm.executeWriteMany(self.corpdb, self.dbCursor, sql, newRows, writeCursor=self.dbConn.cursor(), charset=self.encoding, use_unicode=self.use_unicode)


    def addLDAMessages(self, ldaStatesFile):
        """Creates a LDA topic version of message table

        Parameters
        ----------
        ldaStatesFile : str
            Path to file created with addMessageID.py

        Returns
        -------
        tableName : str
            Name of LDA message table: corptable_lda$ldaStatesFileBaseName
        """
        fin = open(ldaStatesFile, 'r') #done first so throws error if not existing
        baseFileName = os.path.splitext(os.path.basename(ldaStatesFile))[0].replace('-', '_')
        tableName = "%s_lda$%s" %(self.corptable, baseFileName)

        #Create Table:
        columnNames, messageIndex, messageIdIndex = self.createTable(tableName, modify='LONGTEXT')

        commentLine = re.compile('^\#')
        ldaColumnLabels = ['doc', 'message_id', 'index', 'term_id', 'term', 'topic_id']
        labelRange = list(range(len(ldaColumnLabels)))
        ldas = dict() #stored ldas currently being looked at
        msgsAtTime = 100
        msgsWritten = 0

        ##iterate through file:
        for line in fin:
            if not commentLine.match(line):
                line = line.strip()
                fields = line.split()
                currentLDA = dict( [(ldaColumnLabels[i], fields[i]) for i in  labelRange])
                currentId = currentLDA['message_id']
                if currentId not in ldas:
                    if  len(ldas) >= msgsAtTime:#insert
                        self.insertLDARows(ldas, tableName, columnNames, messageIndex, messageIdIndex)
                        ldas = dict() #reset memory
                        msgsWritten += msgsAtTime
                        if msgsWritten % 20000 == 0:
                            dlac.warn("  %.1fk messages' lda topics written" % (msgsWritten/float(1000)))
                    ldas[currentId] = [currentLDA]
                else:
                    ldas[currentId].append(currentLDA)

        #write remainder:
        self.insertLDARows(ldas, tableName, columnNames, messageIndex, messageIdIndex)

        #re-enable keys:
        mm.enableTableKeys(self.corpdb, self.dbCursor, tableName, charset=self.encoding, use_unicode=self.use_unicode)

        return tableName

    def addTokenizedMessages(self):
        """Creates a parsed version of the message table

        Returns
        -------
        tableName : str
            Name of tokenized message table: corptable_tok.
        """
        tableName = "%s_tok" %(self.corptable)
        tokenizer = Tokenizer(use_unicode=self.use_unicode)

        #Create Table:
        columnNames, messageIndex, messageIdIndex = self.__createTable(tableName, modify='LONGTEXT')

        #find all groups
        cfRows = self.__findAllGroups()

        #iterate through groups in chunks

        if self.correl_field == 'message_id':#more at a time when messages:
            self.groupsAtTime = 2000
        groupsWritten = 0
        for groups in dlac.chunks(cfRows, self.groupsAtTime):

            #get msgs for groups:
            rows = self.__getMsgsForGroups(groups, columnNames, messageIndex)
            messages = [r[messageIndex] for r in rows]

            if messages:
                #tokenize msgs:
                parses = [json.dumps(tokenizer.tokenize(m)) for m in messages]
                self.__writeMsgsForGroups(rows, parses, messageIndex, tableName, columnNames)

                groupsWritten += self.groupsAtTime
                if groupsWritten % 100 == 0:
                    dlac.warn("  %.1fk %ss' messages tagged and written" % (groupsWritten/float(1000), self.correl_field))
            else:
                dlac.warn("   Warning: No messages for:" + str(groups))

        #re-enable keys:
        mm.enableTableKeys(self.corpdb, self.dbCursor, tableName, charset=self.encoding, use_unicode=self.use_unicode)

        return tableName

    def addSegmentedMessages(self, model="ctb", tmpdir="/tmp"):
        """Exports the messages to a csv, writes to a tmp file, segments, cleans up and reimports as JSON list

        Parameters
        -------
        model : :obj:`str`, optional
            model used for segmentation: ctb (Penn Chinese Treebank) or pku (Beijing Univ.)
        tmpdir : :obj:`dict`, optional
            temp directory for storing intermediate results
        """

        assert model.lower() in ["ctb", "pku"], "Available models for segmentation are CTB or PKU"
        sql = "select %s, %s from %s" % (self.messageid_field, self.message_field, self.corptable)
        rows = mm.executeGetList(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode)

        tmpfile = tmpdir+"/tmpChineseUnsegmented.txt"
        tmpfile_seg = tmpdir+"/tmpChineseSegmented.txt"

        with open(tmpfile, "w+") as a:
            w = csv.writer(a)
            w.writerows(rows)

        os.system("%s %s %s UTF-8 0 > %s" % (dlac.DEF_STANFORD_SEGMENTER , model.lower(), tmpfile, tmpfile_seg))

        new_rows = []
        raw = []
        with open(tmpfile_seg, "r") as b:
            raw = [r for r in b]
        try:
            r = csv.reader(raw)
            new_rows = [i for i in r]
        except:
            new_rows = []
            for row in raw:
                r = csv.reader([row])
                new_rows.append(next(r))

        new_rows = [[i[0].strip().replace(" ",""), # Message_ids shouldn't get split
                     i[1].strip().replace("[ ","[").replace(" ]", "]").replace(" http : //", " http://").replace(" https : //", " https://")]
                    for i in new_rows[:]]

        # os.system("rm %s %s" % (tmpfile, tmpfile_seg))

        new_rows = [(i[0], json.dumps(i[1].split(" "))) for i in new_rows[:]]

        # Now that we have the new rows, we should insert them. 1) Create table, 2) insert
        tableName = self.corptable+"_seg"
        sql = "SELECT column_name, column_type FROM INFORMATION_SCHEMA.COLUMNS "
        sql += "WHERE table_name = '%s' AND COLUMN_NAME in ('%s', '%s') and table_schema = '%s'" % (
            self.corptable, self.message_field, self.messageid_field, self.corpdb)
        types = {k:v for k,v in mm.executeGetList(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode)}
        sql2 = "CREATE TABLE %s (" % (self.corptable+"_seg")
        sql2 += "%s %s primary key, %s %s " % (self.messageid_field,
                                                 types[self.messageid_field],
                                                 self.message_field,
                                                 types[self.message_field],
                                                 )
        sql2 += ")"
        mm.execute(self.corpdb, self.dbCursor, "drop table if exists "+tableName, charset=self.encoding, use_unicode=self.use_unicode)
        mm.execute(self.corpdb, self.dbCursor, sql2, charset=self.encoding, use_unicode=self.use_unicode)
        mm.standardizeTable(self.corpdb, self.dbCursor, tableName, collate=dlac.DEF_COLLATIONS[self.encoding.lower()], engine=dlac.DEF_MYSQL_ENGINE, charset=self.encoding, use_unicode=self.use_unicode)
        alter = """ALTER TABLE %s MODIFY %s LONGTEXT""" % (tableName, self.message_field)
        mm.execute(self.corpdb, self.dbCursor, alter, charset=self.encoding, use_unicode=self.use_unicode)

        sql = "INSERT INTO %s " % (tableName)
        sql += " VALUES (%s, %s)"
        N = dlac.MYSQL_BATCH_INSERT_SIZE
        totalLength = len(new_rows)
        for l in range(0, totalLength, N):
            print("Inserting rows (%5.2f%% done)" % (float(min(l+N,totalLength))*100/totalLength))
            mm.executeWriteMany(self.corpdb, self.dbCursor, sql, new_rows[l:l+N], writeCursor=self.dbConn.cursor(), charset=self.encoding)

    def addTweetPOSMessages(self):
        """Creates a POS tagged (by TweetNLP) version of the message table

        Returns
        -------
        tableName : str
            Name of POS message table: corptable_tweetpos
        """
        tableName = "%s_tweetpos" %(self.corptable)
        try:
            tagger = TweetNLP()
        except NameError:
            dlac.warn("Method not available without TweetNLP interface")
            raise

        #Create Table:
        columnNames, messageIndex, messageIdIndex = self.__createTable(tableName)

        #find all groups
        cfRows = self.__findAllGroups()

        #iterate through groups in chunks
        groupsWritten = 0
        for groups in dlac.chunks(cfRows, self.groupsAtTime):

            #get msgs for groups:
            rows = self.__getMsgsForGroups(groups, columnNames, messageIndex)
            messages = [r[messageIndex] for r in rows]

            if messages:
                #tokenize msgs:
                parses = [json.dumps(tagger.tag(m)) for m in messages]
                self.__writeMsgsForGroups(rows, parses, messageIndex, tableName, columnNames)

                groupsWritten += self.groupsAtTime
                if groupsWritten % 100 == 0:
                    dlac.warn("  %.1fk %ss' messages tagged and written" % (groupsWritten/float(1000), self.correl_field))
            else:
                dlac.warn("   Warning: No messages for:" + str(groups))

        #re-enable keys:
        mm.enableTableKeys(self.corpdb, self.dbCursor, tableName, charset=self.encoding, use_unicode=self.use_unicode)

        return tableName

    def addTweetTokenizedMessages(self):
        """Creates a tokenized (by TweetNLP) version of the message table

        Returns
        -------
        tableName : str
            Name of tokenized message table: corptable_tweettok
        """
        tableName = "%s_tweettok" %(self.corptable)
        try:
            tokenizer = TweetNLP()
        except NameError:
            dlac.warn("Method not available without TweetNLP interface")
            raise

        #Create Table:
        columnNames, messageIndex, messageIdIndex = self.__createTable(tableName)

        #find all groups
        cfRows = self.__findAllGroups()

        #iterate through groups in chunks
        groupsWritten = 0
        for groups in dlac.chunks(cfRows, self.groupsAtTime):

            #get msgs for groups:
            rows = self.__getMsgsForGroups(groups, columnNames, messageIndex)
            messages = [r[messageIndex] for r in rows]

            if messages:
                #tokenize msgs:
                parses = [json.dumps(tokenizer.tokenize(m)) for m in messages]
                self.__writeMsgsForGroups(rows, parses, messageIndex, tableName, columnNames)

                groupsWritten += self.groupsAtTime
                if groupsWritten % 100 == 0:
                    dlac.warn("  %.1fk %ss' messages tagged and written" % (groupsWritten/float(1000), self.correl_field))
            else:
                dlac.warn("   Warning: No messages for:" + str(groups))

        #re-enable keys:
        mm.enableTableKeys(self.corpdb, self.dbCursor, tableName, charset=self.encoding, use_unicode=self.use_unicode)

        return tableName

    def addSentTokenizedMessages(self, sentPerRow = False, cleanMessages = None):
        """Creates a sentence tokenized version of message table

        Returns
        -------
        tableName : str
            Name of sentence tokenized message table: corptable_stoks or corptable_sent.
        """
        if sentPerRow:
            tableName = "%s_sent" % (self.corptable)
        else:
            tableName = "%s_stoks" % (self.corptable)
        sentDetector = nltk.data.load('tokenizers/punkt/english.pickle')

        #Create Table:
        if sentPerRow:
            modify = 'VARCHAR(64)'
        else:
            modify = ''
        columnNames, messageIndex, messageIdIndex = self.__createTable(tableName, modify)
        
        if cleanMessages:

            ### get lexical normalization dictionaries
            normalizeDict = {}
            
            # Han, Bo  and  Cook, Paul  and  Baldwin, Timothy, 2012
            sql = """select word, norm from %s.%s""" % (dlac.DEF_LEXICON_DB, "han_bo_emnlp_dict")
            normalizeDict.update(dict(list(mm.executeGetList(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode))))
            
            # Fei Liu, Fuliang Weng, Bingqing Wang, Yang Liu, 2011 
            # Fei Liu, Fuliang Weng, Xiao Jiang, 2012
            sql = """select word, norm from %s.%s""" % (dlac.DEF_LEXICON_DB, "liu_weng_test_set_3802")
            normalizeDict.update(dict(list(mm.executeGetList(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode))))
        

        #find all groups
        cfRows = self.__findAllGroups()

        #iterate through groups in chunks
        groupsWritten = 0
        for groups in dlac.chunks(cfRows, self.groupsAtTime):
            sentRows = list()
            
            #get msgs for groups:
            sql = """SELECT %s from %s where %s IN ('%s')""" % (','.join(columnNames), self.corptable, self.correl_field, "','".join(str(g) for g in groups))
            rows = list(mm.executeGetList(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode))#, False)
            rows = [r for r in rows if r[messageIndex]]
            messages = [r[messageIndex] for r in rows]

            if messages:
                insert_idx_start = 0
                insert_idx_end = dlac.MYSQL_BATCH_INSERT_SIZE
                #tokenize msgs:
                # parses = map(lambda m: json.dumps(sentDetector.tokenize(tc.removeNonAscii(tc.treatNewlines(m.strip())))), messages)
                parses = None
                if self.use_unicode:
                    if cleanMessages:
                        parses = [json.dumps(sentDetector.tokenize((tc.sentenceNormalization(m.strip(), normalizeDict, self.use_unicode)))) for m in messages]
                    else:
                        parses = [json.dumps(sentDetector.tokenize(tc.removeNonUTF8(tc.treatNewlines(m.strip())))) for m in messages]
                        #parses = [json.dumps(sentDetector.tokenize(tc.removeNonUTF8(m.strip()))) for m in messages]
                else:
                    parses = [json.dumps(sentDetector.tokenize(tc.removeNonUTF8(tc.treatNewlines(m.strip())))) for m in messages]
                    #parses = [json.dumps(sentDetector.tokenize(tc.removeNonUTF8(m.strip()))) for m in messages]
                    #parses = [json.dumps(sentDetector.tokenize(tc.removeNonAscii(tc.treatNewlines(m.strip())))) for m in messages]

                #add msgs into new tables
                sql = """INSERT INTO """+tableName+""" ("""+', '.join(columnNames)+\
                        """) VALUES ("""  +", ".join(['%s']*len(columnNames)) + """)"""
                for i in range(len(rows)):
                    rows[i] = list(rows[i])
                    if sentPerRow:
                        for j, parse in enumerate(ast.literal_eval(parses[i]), 1):
                            sentRows.append(list(rows[i]))
                            sentRows[-1][messageIDIndex] = str(rows[i][messageIDIndex]) + "_" + str(j).zfill(2)
                            sentRows[-1][messageIndex] = parse
                    elif i < len(parses):
                        sentRows.append(rows[i])#debug: take out copy if eveyrthing ok to run faster
                        sentRows[i][messageIndex] = str(parses[i])
                    else:
                        dlac.warn("   warning: row %d: %s has no parse; last parse %d: %s" % (i, str(rows[i]), len(parses) - 1, str(parses[-1])))

                        # for s in range(len(rows)):#DEBUG: add back in "copy" row above to see differences in original versus parse
                        #     print("\n%d; mid: %s"%(s, str(rows[s][messageIDIndex])))
                        #     print(rows[s][messageIndex])
                        #     try: 
                        #         print(parses[s])
                        #     except IndexError:
                        #         print("!NO PARSE!")
                        # sys.exit(1)

                while insert_idx_start < len(rows):
                    dataToWrite = sentRows[insert_idx_start:min(insert_idx_end, len(rows))]
                    #_warn("Inserting rows %d to %d... " % (insert_idx_start, insert_idx_end))
                    mm.executeWriteMany(self.corpdb, self.dbCursor, sql, dataToWrite, writeCursor=self.dbConn.cursor(), charset=self.encoding, use_unicode=self.use_unicode)
                    insert_idx_start += dlac.MYSQL_BATCH_INSERT_SIZE
                    insert_idx_end += dlac.MYSQL_BATCH_INSERT_SIZE
                
                groupsWritten += self.groupsAtTime
                if groupsWritten % 10000 == 0:
                    dlac.warn("  %.1fk %ss' messages sent tokenized and written" % (groupsWritten/float(1000), self.correl_field))
            else:
                dlac.warn("   Warning: No messages for:" + str(groups))

        #re-enable keys:
        mm.enableTableKeys(self.corpdb, self.dbCursor, tableName, charset=self.encoding, use_unicode=self.use_unicode)

        return tableName

    def printTokenizedLines(self, filename, whiteListFeatTable = None):
        """Prints tokenized messages in format mallet can use

        Parameters
        ----------
        filename : str
            name of file to print to.
        whiteListFeatTable : :obj:`str`, optional
            name of white list feature table.
        """
        imp.reload(sys)
        if not self.use_unicode: sys.setdefaultencoding('utf8')
        sql = """SELECT %s, %s  from %s""" % (self.messageid_field, self.message_field,self.corptable+'_tok')
        messagesEnc = mm.executeGetList(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode)
        try:
            messagesTok = [(m[0], json.loads(m[1])) for m in messagesEnc]
        except ValueError:
            raise ValueError("One of the tokenized messages was badly JSON encoded, please check your data again. (Maybe MySQL truncated the data?)")

        whiteSet = None
        if whiteListFeatTable:
            sql = "SELECT distinct feat FROM %s " % whiteListFeatTable[0]
            whiteSet = set([s[0] for s in mm.executeGetList(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode)])

        f = open(filename, 'w')
        for m in messagesTok:
            toks = m[1]
            if whiteSet:
                toks = [w for w in toks if w in whiteSet]
            f.write("""%s en %s\n""" %(m[0], ' '.join([s for s in toks])))
        f.close()
        dlac.warn("Wrote tokenized file to: %s"%filename)

    def parseAndWriteMessages(self, sp, tableNames, messages, messageIndex, columnNames, rows):
        """Parses, then write messages, used for parallelizing parsing

        Parameters
        -------
        sp : StanfordParser object
            ?????
        tableNames : dict
            {parse type: tablename}
        messages : list
            messages to be parsed
        messageIndex : int
            index of message field
        columnNames : list
            column names in table
        rows : list
            complete row from mysql table

        Returns
        -------
        True
        """
        parses = sp.parse(messages)
        #add msgs into new tables
        for pt, tableName in list(tableNames.items()):
            sql = """REPLACE INTO """+tableName+""" ("""+', '.join(columnNames)+\
                """) VALUES ("""  +", ".join(['%s']*len(columnNames)) + """)"""
            for i in range(len(rows)):
                rows[i] = list(rows[i])
                rows[i][messageIndex] = str(parses[i][pt])

            mm.executeWriteMany(self.corpdb, self.dbCursor, sql, rows, writeCursor=self.dbConn.cursor(), charset=self.encoding)
        return True
    
    def addParsedMessages(self):
        """Creates a parsed version of the message table

        Returns
        -------
        tableNames : dict
            Dictionary of table names: {"pos": corptable_pos, "const": corptable_const, "dep": corptable_dep}
        """
        parseTypes = ['pos', 'const', 'dep']
        tableNames = dict( [(t, "%s_%s" %(self.corptable, t)) for t in parseTypes] )

        #Create Tables: (TODO make continue)
        for t, name in list(tableNames.items()):
            sql = "CREATE TABLE IF NOT EXISTS %s like %s" % (name, self.corptable)
            mm.execute(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode)
            mm.standardizeTable(self.corpdb, self.dbCursor, name, collate=dlac.DEF_COLLATIONS[self.encoding.lower()], engine=dlac.DEF_MYSQL_ENGINE, charset=self.encoding, use_unicode=self.use_unicode)
            mm.enableTableKeys(self.corpdb, self.dbCursor, name, charset=self.encoding, use_unicode=self.use_unicode)#just incase interrupted, so we can find un-parsed groups

        #Find column names:
        columnNames = list(mm.getTableColumnNameTypes(self.corpdb, self.dbCursor, self.corptable, charset=self.encoding, use_unicode=self.use_unicode).keys())
        messageIndex = columnNames.index(self.message_field)

        #find if parsed table already has rows:
        countsql = """SELECT count(*) FROM %s""" % (tableNames[parseTypes[0]])
        cnt = mm.executeGetList(self.corpdb, self.dbCursor, countsql, charset=self.encoding, use_unicode=self.use_unicode)[0][0]

        #find all groups that are not already inserted
        usql = """SELECT %s FROM %s GROUP BY %s""" % (self.correl_field, self.corptable, self.correl_field)
        if cnt: #limit to those not done yet:
            usql = """SELECT a.%s FROM %s AS a LEFT JOIN %s AS b ON a.%s = b.%s WHERE b.%s IS NULL group by a.%s""" % (
                self.correl_field, self.corptable, tableNames[parseTypes[0]], self.messageid_field, self.messageid_field, self.messageid_field, self.correl_field)
        #msgs = 0#keeps track of the number of messages read
        cfRows = [r[0] for r in mm.executeGetList(self.corpdb, self.dbCursor, usql, charset=self.encoding, use_unicode=self.use_unicode)]
        dlac.warn("parsing messages for %d '%s's"%(len(cfRows), self.correl_field))

        #disable keys (waited until after finding groups)
        for t, name in list(tableNames.items()):
            mm.disableTableKeys(self.corpdb, self.dbCursor, name, charset=self.encoding, use_unicode=self.use_unicode)

        #iterate through groups in chunks
        if not any(field in self.correl_field.lower() for field in ["mess", "msg"]) or self.correl_field.lower().startswith("id"):
            self.groupsAtTime = 10 # if user ids
            dlac.warn("""Parsing at the non-message level is not recommended. Please rerun at message level""", attention=True)

        psAtTime = dlac.CORES / 4

        try:
            sp = StanfordParser()
        except NameError:
            dlac.warn("Method not available without StanfordParser interface")
            raise
        groupsWritten = 0
        activePs = set()
        for groups in dlac.chunks(cfRows, self.groupsAtTime):
            #get msgs for groups:
            sql = """SELECT %s from %s where %s IN ('%s')""" % (','.join(columnNames), self.corptable, self.correl_field, "','".join(str(g) for g in groups))
            rows = list(mm.executeGetList(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode))
            rows = [row for row in rows if row[messageIndex] and not row[messageIndex].isspace()]
            messages = [r[messageIndex] for r in rows]
            messages = [m if m else '_' for m in messages]

            #check for limit of active processes
            while (len(activePs) >= psAtTime):
                time.sleep(10)
                toRemove = set()
                for proc in activePs:
                    if not proc.is_alive():
                        toRemove.add(proc)
                        proc.join()
                        groupsWritten += self.groupsAtTime
                        if groupsWritten % 200 == 0:
                            dlac.warn("  %.1fk %ss' messages parsed and written" % (groupsWritten/float(1000), self.correl_field))
                for proc in toRemove:
                    activePs.remove(proc)
                    dlac.warn (" %s removed. (processes running: %d)" % (str(proc), len(activePs)) )


            groupsWritten += self.groupsAtTime
            if groupsWritten % 200 == 0:
                dlac.warn("  %.1fk %ss' messages parsed and written" % (groupsWritten/float(1000), self.correl_field))

            #parse msgs
            p = multiprocessing.Process(target=MessageTransformer.parseAndWriteMessages, args=(self, sp, tableNames, messages, messageIndex, columnNames, rows))
            dlac.warn (" %s starting. (processes previously running: %d)" % (str(p), len(activePs)) )
            p.start()
            activePs.add(p)

        #wait for remaining processes
        temp = activePs.copy()
        for proc in temp:
            activePs.remove(proc)
            proc.join()

        #re-enable keys:
        for t, name in list(tableNames.items()):
            mm.enableTableKeys(self.corpdb, self.dbCursor, name, charset=self.encoding, use_unicode=self.use_unicode)

        return tableNames



#!/usr/bin/python

__author__ = "Lukasz Dziurzynski"
__copyright__ = "Copyright 2012"
__credits__ = []
__license__ = "Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported License: http://creativecommons.org/licenses/by-nc-sa/3.0/"
__version__ = "0.1"
__maintainer__ = "Lukasz Dziurzynski"
__email__ = "lukaszdz@sas.upenn.edu"


import sys
import time
import argparse
import MySQLdb
import re
import pickle
from pprint import pprint
from math import sqrt, log, isnan
from numpy import array, mean, std, var, isnan, zeros
import numpy.random as rand
from scipy.stats.stats import pearsonr
from scipy.stats import scoreatpercentile
from math import floor, ceil, log
import rpy2.robjects as ro
from rpy2.robjects.packages import importr
import rpy2.rlike.container as rlc
from wwbp import notify
import wwbp.mysqlMethods as mm
import array
from numpy import linspace
try:
    import rpy2.interactive 
    ro.r.library('gregmisc')
except ImportError:
    fwc.warn("rpy2 cannot be imported")
    pass

#DB INFO:
HOST = '127.0.0.1'
USER = os.getlogin()
PASSWD = ''

#plotting:
N_HIST_BREAKS = 16
#http://research.stowers-institute.org/efg/R/Graphics/Basics/mar-oma/index.htm
#http://www.r-bloggers.com/setting-graph-margins-in-r-using-the-par-function-and-lots-of-cow-milk/
#rpy2.interactive.packages.utils.install_packages('gregmisc')

class StatsPlotter(object):
    def __init__(self, heights=700, widths=700):
        self.grdevices = importr('grDevices')
        self.graphics = importr('graphics')
        self.stats = importr('stats')
        self.heights = heights
        self.widths = widths

    def plotDescStats(self, dataDict, filename=None, n_breaks=None, invValueFunc=None):
        """Plot boxplots and histograms for the specified dataDict.
           dataDict maps a string (column name) to a list of its values.
           Returns a dictionary containing descriptive statistics."""
        descStatdict = dict()
        for col in dataDict:
            colDescStats = dict()
            data = [x for x in dataDict[col] if x != None]
            data = list(map(float, data))
            total_count = len(dataDict[col])
            if data:
                colDescStats['avg'] = mean(data)
                colDescStats['count'] = len(data)
                colDescStats['std'] = std(data)
                colDescStats['min'] = max(data)
                colDescStats['max'] = min(data)
                ro_data = ro.FloatVector(data)
                ro_breaks = 'FD'
                if n_breaks:
                    ro_breaks = ro.FloatVector(list(linspace(min(data), max(data), n_breaks, True)))
                if filename:
                    self.grdevices.png(file="%s_%s_box.png"%(filename, col), width=self.widths, height=self.heights)
                    #ro.r.boxplot(ro_data, main="%s n=%d, N=%d"%(col, len(data), total_count), col='royalblue4', xlab="value range (in [%2.2f, %2.2f])"%(min(data), max(data)))
                    ro.r.boxplot(ro_data, main="%s N=%d/%d"%(col, len(data), total_count), col='royalblue4', xlab="value range (in [%2.2f, %2.2f])"%(min(data), max(data)))
                    self.grdevices.dev_off()
                    self.grdevices.png(file="%s_%s_hist.png"%(filename, col), width=self.widths, height=self.heights)
                    xaxt = 's'
                    if invValueFunc:
                        xaxt='n'
                    #ro.r.hist(ro_data, breaks=ro_breaks, col='royalblue4', main="%s n=%d, N=%d"%(col, len(data), total_count), xlab="value range (in [%2.2f, %2.2f])"%(min(data), max(data)))
                    ro.r.hist(ro_data, breaks=ro_breaks, col='royalblue4', main="%s N=%d/%d"%(col, len(data), total_count), xlab="value range (in [%2.2f, %2.2f])"%(min(data), max(data)), xaxt=xaxt)
                    if invValueFunc:
                        ro.r.axis(1, at=ro.r.axTicks(1), labels=ro.FloatVector(list(map(invValueFunc, list(ro.r.axTicks(1))))))
                    self.grdevices.dev_off()
            descStatdict[col] = colDescStats
        return descStatdict

    def plotPie(self, dataDict, filename=None, colors=None, labels=None, startAngle=0):
        """Plot a pie chart for the specified dataDict.
           dataDict maps a string (column name) to a list of its values."""
        for col in dataDict:
            data = [x for x in dataDict[col] if x != None]
            data = list(map(float, data))
            #total_count = len(dataDict[col])
            if data:
                ro_data = ro.StrVector(data)
                ro_frequencies = ro.r.table(ro_data)
                ro_percentages = ro.FloatVector( ro.r.round( (ro_frequencies.ro / ro.r.sum(ro_frequencies)).ro * 100.0, 2) )
                ro_labels = ro.r.names(ro_frequencies)
                ro_colors = self.grdevices.rainbow(len(ro_frequencies))
                if labels:
                    ro_labels = ro.StrVector(labels)
                ro_labels = ro.r.paste(ro_labels, ' ', ro_percentages, '%', sep="")
                if colors:
                    ro_colors = ro.StrVector(colors)
                if filename:
                    self.grdevices.png(file="%s_%s_pie.png"%(filename, col), width=self.widths, height=self.heights)
                    ro.r.pie(ro_frequencies, main="%s N=%d"%(col, len(data)), col=ro_colors, labels=ro_labels, **{'init.angle':startAngle } )
                    self.grdevices.dev_off()        

    def plotScatter(self, x_name, x_values, y_name, y_values, filename=None):
        """Make a scatterplot"""
        ro_x = ro.FloatVector(x_values)
        ro_y = ro.FloatVector(y_values)
        if filename:
            self.grdevices.png(file="%s.png"%(filename), width=self.widths, height=self.heights)
        else:
            self.grdevices.png(file="scat_%s-%s.png"%(x_name, y_name), width=self.widths, height=self.heights)
        ro.globalenv['x'] = ro_x
        ro.globalenv['y'] = ro_y
        ro.r('regr <- lm(y~x)')
        intercept = ro.r('int <- regr$coefficients[1]')[0]
        beta = ro.r('beta <- regr$coefficients[2]')[0]
        r_squared = ro.r('r_squared <- summary(regr)$r.squared')[0]
        ro.r.plot(ro_x, ro_y, xlab=x_name, ylab=y_name, main="Y ~ %.4g + %.4g x   r2: %2.4f"%(intercept, beta, r_squared))
        ro.r('abline(regr, col="red")')
        if filename:
            self.grdevices.dev_off()

    def _getBins(self, value_list):
        n = len(value_list)
        if n < 200:
            return ceil( log(n,2) + 1 )
        else:
            p75 = scoreatpercentile(value_list, 75)
            p25 = scoreatpercentile(value_list, 25)
            return ceil( (max(value_list)-min(value_list)) / (2.0*(p75-p25)*(n**(-1.0/3.0))) )
        

    def plot2dHist(self, x_name, x_values, y_name, y_values, filename=None):
        """Plot the 2d histogram from gregmisc R package"""
        ro_x = ro.FloatVector(x_values)
        ro_y = ro.FloatVector(y_values)
        if filename:
            self.grdevices.png(file="%s.png"%(filename), width=self.widths, height=self.heights)
        else:
            self.grdevices.png(file="2dHist_%s-%s.png"%(x_name, y_name), width=self.widths, height=self.heights)
        ro.globalenv['x'] = ro_x
        ro.globalenv['y'] = ro_y
        ro.r('regr <- lm(y~x)')
        intercept = ro.r('int <- regr$coefficients[1]')[0]
        beta = ro.r('beta <- regr$coefficients[2]')[0]
        r_squared = ro.r('r_squared <- summary(regr)$r.squared')[0]
        n_bins = self._getBins(x_values) * self._getBins(y_values)
        rgb_palette = self.grdevices.colorRampPalette(ro.StrVector(['lightyellow', 'orange']), space='rgb')
        ro_colors = ro.StrVector(['white'] + list(rgb_palette(20)))
        ro.r.hist2d(ro_x, ro_y, xlab=x_name, ylab=y_name, main="Y ~ %.4g + %.4g x   r2: %2.4f"%(intercept, beta, r_squared), col=ro_colors, nbins=n_bins)
        ro.r('abline(regr, col="black")')
        if filename:
            self.grdevices.dev_off()

    def plot2dHistGeneralized(self, names_to_x, names_to_y, filename=None):
        """Plot the 2d histogram for all combinations of x and y"""
        n_rows = len(names_to_x)
        n_cols = len(names_to_y)

        if filename:
            self.grdevices.png(file="%s_%s_%s_2dhist.png"%(filename, '-'.join(list(names_to_x.keys())), '-'.join(list(names_to_y.keys())) ), width=200*n_rows, height=400*n_cols)
            ro.r.par(mfrow=ro.IntVector([n_rows, n_cols]))

        for x_name in names_to_x:
            for y_name in names_to_y:
                self.plot2dHist(x_name, names_to_x[x_name], y_name, names_to_y[y_name])
                
        if filename:
            self.grdevices.dev_off()

    def getFloatColStats(self, db, dbCursor, tableName, colsOfNote, filename=None):
        #1. Get column names
        sql = "SELECT column_name from information_schema.columns where table_name='%s'"%tableName
        row = mm.executeGetList(db, dbCursor, sql)
        colNames = []
        for col in colsOfNote:
            colNames.append(row[col][0])

        ncols = len(colsOfNote)

        #. Get users who have at least one message
        sql = "SELECT DISTINCT group_id FROM feat$1gram$messages$user_id$16to16"
        rows = mm.executeGetList(db, dbCursor, sql)
        user_ids = []
        for row in rows:
            user_ids.append(row[0])
        sql_user_ids = ",".join(map(str,user_ids))

        #. Get total row count
        total_count = len(user_ids)

        #. Assemble storage data structure
        dataHolder = []
        for col in colsOfNote:
            dataHolder.append([None]*total_count)

        #. Pull data of interest; use offset if needed
        sql = "SELECT * FROM %s WHERE user_id IN (%s)"%(tableName, sql_user_ids)
        rows = mm.executeGetList(db, dbCursor, sql)
        ii = 0
        for row in rows:
            jj = 0
            for col in colsOfNote:
                dataHolder[jj][ii] = row[col]
                jj += 1
            ii += 1

        #. Link the data to their names
        dataDict = dict()
        for cc in range(ncols):
            dataDict[colNames[cc]] = dataHolder[cc]

        return self.plotDescStats(dataDict, total_count, filename)
        
    def getCategoricalColStats(self, db, dbCursor, tableName, colsOfNote, filename=None):
        #1. Get column names
        sql = "SELECT column_name from information_schema.columns where table_name='%s'"%tableName
        row = mm.executeGetList(db, dbCursor, sql)
        colNames = []
        for col in colsOfNote:
            colNames.append(row[col][0])

        ncols = len(colsOfNote)

        #. Get users who have at least one message
        sql = "SELECT DISTINCT group_id FROM feat$1gram$messages$user_id$16to16"
        rows = mm.executeGetList(db, dbCursor, sql)
        user_ids = []
        for row in rows:
            user_ids.append(row[0])
        sql_user_ids = ",".join(map(str,user_ids))

        #. Get total row count
        total_count = len(user_ids)

        #. Assemble storage data structure
        dataHolder = dict()
        for col in colNames:
            dataHolder[col] = 0
        dataHolder["none specified"] = 0

        #. Pull data of interest
        sql = "SELECT * FROM %s WHERE user_id IN (%s)"%(tableName, sql_user_ids)
        rows = mm.executeGetList(db, dbCursor, sql)
        ii = 0
        for row in rows:
            jj = 0
            has_specified_value = False
            for col in colsOfNote:
                if row[col]:
                    dataHolder[colNames[jj]] += 1
                    if has_specified_value:
                        raise Exception("incorrect assumption; more than one category allowed")
                    has_specified_value = True
                jj += 1
            if not has_specified_value:
                dataHolder["none specified"] += 1
            ii += 1

        #. Calculate descriptive statistics and create plots
        labels = list(dataHolder.keys())
        counts = list(dataHolder.values())
        ro_labels = ro.StrVector(labels)
        ro_counts = ro.IntVector(counts)
        if filename:
            self.grdevices.png(file="%s_hist_cats.png"%(filename), width=self.widths, height=self.heights)
            self.graphics.par(las=2, mar=[5.1, 7.1, 4.1, 2.1])
            ro.r.barplot(ro_counts, main = "Category Histogram, N=%d"%(total_count), beside=True, horiz=True, col='royalblue4', **{"names.arg":ro_labels})
            self.grdevices.dev_off()

if __name__=="__main__":
    sp = StatsPlotter()
    #floatCols = [2, 3, 4] + range(6,14) + [23] + [28]
    #prefix = '600_'
    #sp.getFloatColStats("userstats_en", floatCols, "plots/%sdesc"%prefix)
    #sp.getCategoricalColStats("userstats_en", range(14, 23), "plots/%srelnbins"%prefix)
    (conn, cur, dcur) = mm.dbConnect('fb20')
    #sp.getCategoricalColStats('fb20', cur, "userstats_en",  range(24, 28), '/data/ml/plots/fb20/age_category')
    #sp.getCategoricalColStats('fb20', cur, "userstats_en",  range(14, 23), '/data/ml/plots/fb20/reln_category')
    # N=100000
    # d1 = list(rand.normal(0,2,N))
    # d2 = list(rand.normal(0,1,N))
    # d3 = list(rand.normal(0,17,N))
    # d_all = {"d1":d1, "d2":d2, "d3":d3}
    # e1 = list(rand.exponential(2,N))
    # e2 = list(rand.exponential(14,N))
    # e_all = {'e1':e1, 'e2':e2}
    # sp.plot2dHist('d1', d1, 'd2', d2)
    # sp.plot2dHistGeneralized(d_all, e_all, 'plots/samba')
    mm.warn("descStats.py exits with success :)")
